import { Component, Input, Output, EventEmitter, ViewChildren, QueryList, DoCheck } from "@angular/core";

import { ParamRadioConfig } from "../../formulaire/ngparam";
import { FieldSet } from "../../formulaire/fieldset";
import { ParamFieldLineComponent } from "../param-field-line/param-field-line.component";
import { Field } from "../../formulaire/field";
import { InputField } from "../../formulaire/input-field";
import { SelectField } from "../../formulaire/select-field";
import { SelectFieldModel } from "../../formulaire/select-field-model";
import { FormulairePab } from "../../formulaire/definition/concrete/form-pab";
import { SelectModelFieldLineComponent } from "../select-model-field-line/select-model-field-line.component";
import { FieldsetContainer } from "../../formulaire/fieldset-container";

@Component({
    selector: "field-set",
    templateUrl: "./field-set.component.html",
    styleUrls: ["./field-set.component.scss"]
})
export class FieldSetComponent implements DoCheck {

    @Input()
    private set fieldSet(fs: FieldSet) {
        this._fieldSet = fs;
    }

    public get fields() {
        return this._fieldSet.kids;
    }

    /* public set fieldsetNumber(n: number) {
        this._fieldSet.labelNumber = n;
    } */

    public get title(): string {
        if (! this._fieldSet) {
            return "fs undefined";
        }
        if (! this._fieldSet.label) {
            return "label undefined";
        }
        return this._fieldSet.label;
    }

    public get isValid() {
        return this._isValid;
    }

    /** flag d'affichage des boutons ajouter, supprimer, monter, descendre */
    public get showButtons() {
        return (this._fieldSet.parent instanceof FieldsetContainer);
    }

    /** flag d'activation du bouton monter */
    public get enableUpButton() {
        if (this._fieldSet.parent instanceof FieldsetContainer) {
            return this._fieldSet.parent.getFieldsetPosition(this._fieldSet) !== 0;
        }
        return false;
    }

    /** flag d'activation du bouton descendre */
    public get enableDownButton() {
        if (this._fieldSet.parent instanceof FieldsetContainer) {
            return this._fieldSet.parent.getFieldsetPosition(this._fieldSet) < this._fieldSet.parent.fieldsets.length - 1;
        }
        return false;
    }

    /** flag d'activation du bouton supprimer */
    public get enableRemoveButton() {
        if (this._fieldSet.parent instanceof FieldsetContainer) {
            return this._fieldSet.parent.fieldsets.length > 0;
        }
        return false;
    }

    /**
    * field set attribute
    */
    private _fieldSet: FieldSet;

    @ViewChildren(ParamFieldLineComponent)
    private _paramComponents: QueryList<ParamFieldLineComponent>;

    @ViewChildren(SelectModelFieldLineComponent)
    private _selectModelComponents: QueryList<SelectModelFieldLineComponent>;

    /**
     * événement de changement de validité
     */
    @Output()
    private validChange = new EventEmitter();

    /**
     * événement de changement de valeur d'un input
     */
    @Output()
    private inputChange = new EventEmitter();

    /**
     * événement de demande d'ajout d'un fieldset (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private addFieldset = new EventEmitter();

    /**
     * événement de demande de remontée d'un fieldset dans la liste (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private moveFieldsetUp = new EventEmitter();

    /**
     * événement de demande de descente d'un fieldset dans la liste (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private moveFieldsetDown = new EventEmitter();

    /**
     * événement de demande de suppression d'un fieldset (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private removeFieldset = new EventEmitter();

    /**
     * flag de validité de la saisie
     */
    private _isValid = false;

    /**
     * événement de changement d'état d'un radio
     */
    // tslint:disable-next-line:no-output-on-prefix
    @Output()
    private radio = new EventEmitter<any>();

    /** événement signalant un appui sur TAB ou SHIFT+TAB */
    @Output()
    protected tabPressed = new EventEmitter<any>();

    private hasRadioFix(): boolean {
        if (this._fieldSet.hasInputs) {
            switch (this._fieldSet.getInput(0).radioConfig) {
                case ParamRadioConfig.FIX:
                    return false;

                default:
                    return true;
            }
        }
        return false;
    }

    private hasRadioVar(): boolean {
        if (this._fieldSet.hasInputs) {
            switch (this._fieldSet.getInput(0).radioConfig) {
                case ParamRadioConfig.VAR:
                case ParamRadioConfig.CAL:
                    return true;

                default:
                    return false;
            }
        }
        return false;
    }

    private hasRadioCal(): boolean {
        if (this._fieldSet.hasInputs) {
            switch (this._fieldSet.getInput(0).radioConfig) {
                case ParamRadioConfig.CAL:
                    return true;

                default:
                    return false;
            }
        }
        return false;
    }

    /**
     * détermine si un Field est du type InputField
     */
    private isInputField(f: Field): boolean {
        return f instanceof InputField && f.isDisplayed;
    }

    /**
     * détermine si un Field est du type SelectField
     */
    private isSelectField(f: Field): boolean {
        return (
            f instanceof SelectField
            && ! (f.parentForm instanceof FormulairePab)
            && f.isDisplayed
        );
    }

    /**
     * détermine si un Field est du type SelectFieldModel
     */
    private isSelectModelField(f: Field): boolean {
        return (
            f instanceof SelectFieldModel
            && f.parentForm instanceof FormulairePab
            && (
                f.id === (f.parentForm as FormulairePab).modeleCloisonsSelectId
                || f.id === (f.parentForm as FormulairePab).modeleCloisonAvalSelectId
            )
        );
    }
    /*
     * gestion des événements clic sur les radios :
     * réception d'un message du composant enfant (param-field)
     * cf. https://angular.io/guide/component-interaction#parent-listens-for-child-event
     */
    private onRadioClick(info: any) {
        // on renvoie l'info au parent
        this.radio.emit(info);
    }

    /**
     * calcul de la validité de tous les ParamFieldLineComponent et tous les
     * SelectModelFieldLineComponent de la vue
     */
    private updateValidity() {
        this._isValid = false;
        let paramsAreValid = true;
        let selectAreValid = true;

        if (this._paramComponents) {
            paramsAreValid = false;
            paramsAreValid = this._paramComponents.reduce(
                // callback
                (
                    // accumulator (valeur précédente du résultat)
                    acc,
                    // currentValue (élément courant dans le tableau)
                    param,
                    // currentIndex (indice courant dans le tableau)
                    currIndex,
                    // array (tableau parcouru)
                    array
                ) => {
                    return acc && param.isValid;
                }
                // valeur initiale
                , true);
        }

        if (this._selectModelComponents) {
            selectAreValid = false;
            selectAreValid = this._selectModelComponents.reduce(
                // callback
                (
                    // accumulator (valeur précédente du résultat)
                    acc,
                    // currentValue (élément courant dans le tableau)
                    select,
                    // currentIndex (indice courant dans le tableau)
                    currIndex,
                    // array (tableau parcouru)
                    array
                ) => {
                    return acc && select.isValid;
                }
                // valeur initiale
                , true);
        }

        // global validity
        this._isValid = (paramsAreValid && selectAreValid);

        this.validChange.emit();
    }

    /**
     * Renvoie l'événement au composant du dessus
     */
    public onTabPressed(event) {
        this.tabPressed.emit(event);
    }

    public ngDoCheck() {
        this.updateValidity();
    }

    /**
     * réception d'un événement de validité de ParamFieldLineComponent
     */
    private onParamLineValid(event: boolean) {
        this.updateValidity();
    }

    /**
    * réception d'un événement de changement de valeur d'un input
    */
    private onInputChange(event: boolean) {
        this.inputChange.emit();
    }

    /**
     * relit les valeurs dans l'interface et met à jour les NgParameter
     */
    public updateParametersFromUI() {
        this._paramComponents.forEach(fsc => fsc.updateParameterFromUI());
    }

    /**
     * met à jour les paramètres liés
     */
    public updateLinkedParameters() {
        this._paramComponents.forEach(fsc => fsc.updateLinkedParameter());
    }

    /**
     * clic sur le bouton ajouter
     */
    private onAddClick() {
        this.addFieldset.emit(this._fieldSet);
    }

    /**
     * clic sur le bouton supprimer
     */
    private onRemoveClick() {
        this.removeFieldset.emit(this._fieldSet);
    }

    /**
     * clic sur le bouton monter
     */
    private onMoveUpClick() {
        this.moveFieldsetUp.emit(this._fieldSet);
    }

    /**
     * clic sur le bouton descendre
     */
    private onMoveDownClick() {
        this.moveFieldsetDown.emit(this._fieldSet);
    }
}
