import { Component, ViewChild, AfterContentInit, OnInit } from "@angular/core";

import { Observer } from "jalhyd";

import { GraphTypeSelectComponent } from "./graph-type.component";
import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";
import { I18nService } from "../../services/internationalisation/internationalisation.service";
import { PlottableData } from "../../results/plottable-data";
import { GraphType } from "../../results/graph-type";

@Component({
    selector: "results-graph",
    templateUrl: "./results-graph.component.html",
    styleUrls: [
        "./results-graph.component.scss"
    ]
})
export class ResultsGraphComponent implements AfterContentInit, Observer {
    private _results: PlottableData;

    /** used to briefly destroy/rebuild the chart component, to refresh axis labels (@see bug #137) */
    public displayChart = true;

    @ViewChild(GraphTypeSelectComponent)
    private _graphTypeComponent: GraphTypeSelectComponent;

    /*
     * config du graphe
     */
    public graph_type: string;
    public graph_data = {};
    public graph_options = {
        responsive: true,
        maintainAspectRatio: true,
        animation: {
            duration: 0
        },
        legend: {
            display: false
        },
        title: {
            display: true,
            text: ""
        },
        elements: {
            line: {
                tension: 0
            }
        }
    };

    /** tracks the fullscreen state */
    public get isFullscreen() {
        return (document["fullscreenElement"] !== null);
    }

    public constructor(
        private appSetup: ApplicationSetupService,
        private intlService: I18nService
    ) { }

    public set results(r: PlottableData) {
        this.forceRebuild();
        this._results = r;
        if (this._results && this._graphTypeComponent) {
            this._graphTypeComponent.selectedValue = r.graphType;
        }
    }

    public get availableChartAxis() {
        if (this._results) {
            return this._results.getAvailableChartAxis();
        }
    }

    public get chartX() {
        if (this._results) {
            return this._results.chartX;
        }
    }

    public set chartX(X) {
        if (X !== this.chartX) {
            this._results.chartX = X;
            this.forceRebuild();
            // refresh chart
            this.updateView();
        }
    }

    public get chartY() {
        if (this._results) {
            return this._results.chartY;
        }
    }

    public set chartY(Y) {
        if (Y !== this.chartY) {
            this._results.chartY = Y;
            this.forceRebuild();
            // refresh chart
            this.updateView();
        }
    }

    /**
     * Returns a human readable description of any param / result symbol
     */
    public getChartAxisLabel(symbol: string) {
        return this._results.getChartAxisLabel(symbol);
    }

    public get uitextSelectX() {
        return this.intlService.localizeText("INFO_PARAMFIELD_GRAPH_SELECT_X_AXIS");
    }

    public get uitextSelectY() {
        return this.intlService.localizeText("INFO_PARAMFIELD_GRAPH_SELECT_Y_AXIS");
    }

    public updateView() {
        // (re)generate chart
        switch (this._graphTypeComponent.selectedValue) {
            case GraphType.Histogram:
                this.graph_type = "bar";
                this.generateBarGraph();
                break;

            default:
                this.graph_type = "scatter";
                this.generateScatterGraph();
                break;
        }
    }

    public ngAfterContentInit() {
        this._graphTypeComponent.addObserver(this);
    }

    /** forces Angular to rebuild the chart @see bug #137 */
    private forceRebuild() {
        this.displayChart = false;
        const that = this;
        setTimeout(() => { // trick
            that.displayChart = true;
        }, 10);
    }

    /**
     * génère les données d'un graphe de type "bar"
     */
    private generateBarGraph() {
        const labs = [];
        const dat = [];
        const xSeries = this._results.getValuesSeries(this.chartX);
        const ySeries = this._results.getValuesSeries(this.chartY);
        // both series are supposed to be the same length
        for (let i = 0; i < xSeries.length; i++) {
            labs.push(xSeries[i]);
            dat.push(ySeries[i]);
        }

        const nDigits = this.appSetup.displayDigits;
        this.graph_options["scales"] = {
            xAxes: [{
                gridLines: {
                    offsetGridLines: true
                },
                ticks: {
                    precision: nDigits,
                    callback: function(value, index, values) {
                        return Number(value).toFixed(nDigits);
                    }
                },
                scaleLabel: {
                    display: true,
                    labelString: this.chartX
                }
            }],
            yAxes: [{
                scaleLabel: {
                    display: true,
                    labelString: this.chartY
                }
            }]
        };
        this.graph_options["tooltips"] = {
            callbacks: {
                label: function(tooltipItem, data) {
                    return Number(tooltipItem.yLabel).toFixed(nDigits);
                }
            }
        };

        this.graph_data = {
            labels: labs,
            datasets: [{
                label: "",
                data: dat
            }]
        };
    }

    /**
     * génère les données d'un graphe de type "scatter"
     */
    private generateScatterGraph() {
        const dat = [];
        const xSeries = this._results.getValuesSeries(this.chartX);
        const ySeries = this._results.getValuesSeries(this.chartY);
        // both series are supposed to be the same length
        for (let i = 0; i < xSeries.length; i++) {
            dat.push({
                x: xSeries[i],
                y: ySeries[i]
            });
        }

        const nDigits = this.appSetup.displayDigits;
        this.graph_options["scales"] = {
            xAxes: [{
                type: "linear",
                position: "bottom",
                ticks: {
                    precision: nDigits
                },
                scaleLabel: {
                    display: true,
                    labelString: this.chartX
                }
            }],
            yAxes: [{
                type: "linear",
                position: "left",
                ticks: {
                    precision: nDigits
                },
                scaleLabel: {
                    display: true,
                    labelString: this.chartY
                }
            }]
        };
        this.graph_options["tooltips"] = {
            callbacks: {
                label: function(tooltipItem, data) {
                    return "(" + Number(tooltipItem.xLabel).toFixed(nDigits) + ", " + Number(tooltipItem.yLabel).toFixed(nDigits) + ")";
                }
            }
        };

        this.graph_data = {
            datasets: [{
                label: "",
                data: dat,
                borderColor: "#808080", // couleur de la ligne
                backgroundColor: "rgba(0,0,0,0)",  // couleur de remplissage sous la courbe : transparent
                showLine: "true"
            }]
        };
    }

    public setFullscreen(element) {
        element.requestFullscreen();
    }

    public exitFullscreen() {
        document.exitFullscreen();
    }

    public exportAsImage(element: HTMLDivElement) {
        const canvas: HTMLCanvasElement = element.querySelector("canvas");
        canvas.toBlob((blob) => {
            saveAs(blob, "chart.png");
        }); // defaults to image/png
    }

    // interface Observer

    update(sender: any, data: any) {
        if (sender instanceof GraphTypeSelectComponent) {
            if (data.action === "graphTypeChanged") {
                this._results.graphType = data.value;
            }
        }
        this.updateView();
    }
}
