import { Structure, Nub, ParallelStructure, StructureProperties, Props, Session } from "jalhyd";

import { FormComputeParallelStructures } from "../form-compute-parallel-structures";
import { FormResultFixedVar } from "../form-result-fixedvar";
import { FieldsetContainer } from "../../fieldset-container";
import { FieldSet } from "../../fieldset";
import { SelectField } from "../../select-field";
import { NgParameter } from "../../ngparam";
import { FieldsetTemplate } from "../../fieldset-template";
import { FormulaireNode } from "../../formulaire-node";
import { FormulaireBase } from "./form-base";

export class FormulaireParallelStructure extends FormulaireBase {

    /**
     * id du select configurant le type d'ouvrage
     */
    private __ouvrageSelectId: string;

    constructor() {
        super();
        this._formResult = new FormResultFixedVar(this);

        // remove obsolete observer set by super()
        this.removeObserver(this._formCompute);
        this._formCompute = new FormComputeParallelStructures(this, (this._formResult as FormResultFixedVar));
    }

    private createStructNub(templ: FieldsetTemplate): Nub {
        // !!! attention !!!
        // Il doit y avoir cohérence dans le fichier de conf entre les valeurs defaultXXX et les valeurs possibles pour les select
        // cad valeur par défaut du 1er select (type d'ouvrage), du 2ème (loi de débit).
        // A terme, il faudrait analyser le fichier de conf (dépendances d'existence) pour déterminer automatiquement ces valeurs

        const params = {};
        params["calcType"] = templ.calcTypeFromConfig;
        params["nodeType"] = templ.defaultNodeTypeFromConfig;
        params["structureType"] = templ.defaultStructTypeFromConfig;
        params["loiDebit"] = templ.defaultLoiDebitFromConfig;

        return this.createStructure(new Props(params));
    }

    /**
     * ajoute un Nub Structure
     * @param st structure à ajouter
     * @param after position après laquelle insérer la structure, à la fin sinon
     */
    private addStructureNub(st: Structure, after?: number) {
        this.parallelStructureNub.addChild(st, after);
    }

    private get parallelStructureNub(): ParallelStructure {
        return this.currentNub as ParallelStructure;
    }

    /**
     * Asks JaLHyd to create a Structure Nub as a child of the current Calculator Module
     * and return it; does not store it in the Session (for Structures, not for Calculator Modules)
     * @param p properties for the new Nub
     */
    protected createStructure(p: Props): Structure {
        return Session.getInstance().createNub(p, this.currentNub as ParallelStructure) as Structure;
    }

    /**
     * Replaces the current Nub in the current calculator module, with a new one built with properties "params"
     * @param params properties to build the new Nub (calcType, loiDebit...)
     */
    protected replaceNub(sn: Structure, params: Props): Nub {
        const parent = (this.currentNub as ParallelStructure);
        const newStructure = this.createStructure(params);
        parent.replaceChildInplace(sn, newStructure);
        return newStructure;
    }

    /**
     * Deleted the given child Nub in the current calculator module
     * @param params properties to build the new Nub (calcType, loiDebit...)
     */
    protected deleteNub(sn: Structure) {
        const parent = (this.currentNub as ParallelStructure);
        parent.deleteChild(parent.getIndexForChild(sn));
    }

    public createFieldset(parent: FormulaireNode, json: {}, data?: {}, nub?: Nub): FieldSet {
        if (json["calcType"] === "Structure") {
            // indice après lequel insérer le nouveau FieldSet
            const after = data["after"];

            const res: FieldSet = new FieldSet(parent);
            let sn: Nub;
            if (nub) { // use existing Nub (build interface based on model)
                sn = nub;
            } else {
                sn = this.createStructNub(data["template"]);
                this.addStructureNub(sn as Structure, after);
            }
            res.setNub(sn, false);

            if (after !== undefined) {
                parent.kids.splice(after + 1, 0, res);
            } else {
                parent.kids.push(res);
            }

            this.resetResults();

            return res;
        } else {
            return super.createFieldset(parent, json, data);
        }
    }

    protected parseOptions(json: {}) {
        super.parseOptions(json);

        // id du select configurant le type d'ouvrage
        this.__ouvrageSelectId = this.getOption(json, "ouvrageSelectId");
    }

    public afterParseFieldset(fs: FieldSet) {
        // si le FieldSet contient le select de type d'ouvrage
        if (this.__ouvrageSelectId) {
            const sel = fs.getFormulaireNodeById(this.__ouvrageSelectId);
            if (sel) {
                // on abonne le formulaire aux propriétés du FieldSet
                fs.properties.addObserver(this);
            }
        }
    }

    public moveFieldsetUp(fs: FieldSet) {
        if (fs.nub instanceof Structure) {
            // déplacement du nub
            fs.nub.parent.moveChildUp(fs.nub);
            // déplacement du fieldset
            this.fieldsetContainer.moveFieldsetUp(fs);

            this.resetResults();
        } else {
            super.moveFieldsetUp(fs);
        }
    }

    public moveFieldsetDown(fs: FieldSet) {
        if (fs.nub instanceof Structure) {
            // déplacement du nub
            fs.nub.parent.moveChildDown(fs.nub);
            // déplacement du fieldset
            this.fieldsetContainer.moveFieldsetDown(fs);

            this.resetResults();
        } else { super.moveFieldsetDown(fs); }
    }

    public removeFieldset(fs: FieldSet) {
        if (fs.nub instanceof Structure) {
            // suppression du sous-nub dans le Nub parent
            this.deleteNub(fs.nub);

            // suppression du fieldset
            this.fieldsetContainer.removeFieldset(fs);

            this.resetResults();
        } else { super.removeFieldset(fs); }
    }

    protected completeParse(json: {}) {
        this.subscribeFieldsetContainer();
    }

    private get fieldsetContainer(): FieldsetContainer {
        const n = this.getFormulaireNodeById("struct_container");
        if (n === undefined || !(n instanceof FieldsetContainer)) {
            throw new Error("l'élément 'struct_container' n'est pas du type FieldsetContainer");
        }
        return n as FieldsetContainer;
    }

    /**
     * Après une modification, détermine si les propriétés d'un Fieldset sont compatibles
     * entre elles et les ajuste au besoin
     * @param props propriétés à vérifier
     * @param name nom de la propriété qui vient de changer
     * @param val nouvelle valeur de la propriété
     */
    private adjustProperties(props: Props, name: string, val: any) {
        if (name === "structureType") {
            if (! StructureProperties.isCompatibleValues(
                val, props.getPropValue("loiDebit"), this.currentNub as ParallelStructure
            )) {
                // currentNub should always be a ParallelStructure here
                const ld = StructureProperties.findCompatibleLoiDebit(
                    props.getPropValue("structureType"),
                    [],
                    this.currentNub as ParallelStructure
                );
                props.setPropValue("loiDebit", ld);
            }
        }
    }

    /**
     * abonnement en tant qu'observateur du FieldsetContainer
     */
    private subscribeFieldsetContainer() {
        this.fieldsetContainer.addObserver(this);
    }

    /**
     * abonnement en tant qu'observateur des NgParameter des FieldSet contenus dans le FieldsetContainer
     */
    private subscribeStructureInputFields(fs: FieldSet) {
        for (const n of fs.allFormElements) {
            if (n instanceof NgParameter) {
                n.addObserver(this);
            }
        }
    }
    /**
     * abonnement en tant qu'observateur du SelectField des FieldSet contenus dans le FieldsetContainer
     */
    private subscribeStructureSelectFields(fs: FieldSet) {
        for (const n of fs.allFormElements) {
            if (n instanceof SelectField) {
                n.addObserver(this);
            }
        }
    }

    // interface Observer

    public update(sender: any, data: any) {

        super.update(sender, data);

        if (sender instanceof FieldsetContainer) {
            switch (data.action) {
                case "newFieldset":
                    this.reset();
                    this.subscribeStructureInputFields(data["fieldset"]);
                    this.subscribeStructureSelectFields(data["fieldset"]);
            }
        } else if (sender instanceof FieldSet && data.action === "propertyChange") {
            switch (sender.id) {
                case "fs_ouvrage":
                    const props = sender.properties;
                    // ensure loiDebit is set
                    props.setPropValue("loiDebit", data.value);
                    this.adjustProperties(props, data["name"], data["value"]);
                    // replace Structure Nub
                    const newNub = this.replaceNub((sender.nub as Structure), props);
                    sender.setNub(newNub);
                    // treat the fieldset as new to re-subscribe to Nub properties change events
                    this.afterParseFieldset(sender);
                    this.reset();
                    break;
            }
        }
    }
}
