import { CalculatorType, ComputeNodeType, Nub, Props, Observer, Session, SectionNub, acSection } from "jalhyd";

import { FormulaireElement } from "../formulaire-element";
import { NgParameter, ParamRadioConfig } from "../ngparam";
import { Field } from "../field";
import { StringMap } from "../../stringmap";
import { FormulaireNode } from "../formulaire-node";
import { FieldSet } from "../fieldset";
import { FieldsetContainer } from "../fieldset-container";
import { SelectField } from "../select-field";
import { DeepFieldsetIterator } from "../form-iterator/deep-fieldset-iterator";
import { DeepFormulaireElementIterator } from "../form-iterator/deep-element-iterator";
import { TopFormulaireElementIterator } from "../form-iterator/top-element-iterator";
import { CalculatorResults } from "../../results/calculator-results";
import { ServiceFactory } from "../../services/service-factory";

/**
 * classe de base pour tous les formulaires
 */
export abstract class FormulaireDefinition extends FormulaireNode implements Observer {
    /** nom du module de calcul */
    private _calculatorName: string;

    /** Nub courant */
    protected _currentNub: Nub;

    /** propriétés par défaut (lues si _currentNub === undefined ) */
    protected _props = {};

    /** fichier de configuration */
    private _jsonConfig: {};

    /** aide en ligne */
    private _helpLink: string;

    /** clé-valeurs du fichier de localisation spécifique à ce module */
    private _specificLocalisation: StringMap;

    /** ISO 639-1 language code of the current language (to avoid unnecessary localisation reload) */
    private _currentLanguage: string;

    constructor() {
        super(undefined);
    }

    // surcharge de FormulaireNode::get:uid()
    // l'UID d'un formulaire est l'UID de son Nub !
    public get uid() {
        if (this._currentNub) {
            return this._currentNub.uid;
        } else {
            throw new Error("Aucun Nub associé au formulaire !");
        }
    }

    public get specificLocalisation() {
        return this._specificLocalisation;
    }

    public get currentLanguage() {
        return this._currentLanguage;
    }

    public get calculatorType(): CalculatorType {
        const props = this._currentNub === undefined ? this.defaultProperties : (this._currentNub.properties as Props).props;
        return props["calcType"];
    }

    public get nodeType(): ComputeNodeType {
        const props = this._currentNub === undefined ? this.defaultProperties : (this._currentNub.properties as Props).props;
        return props["nodeType"];
    }

    public get calculatorName() {
        return this._calculatorName;
    }

    public set calculatorName(name: string) {
        this._calculatorName = name;
        this.notifyObservers({
            "action": "nameChanged",
            "name": name
        }, this);
        // convenient trick to notify param-link components
        ServiceFactory.instance.formulaireService.propagateFormNameChange(this, name);
    }

    public get jsonConfig(): {} {
        return this._jsonConfig;
    }

    public get defaultProperties() {
        return this._props;
    }

    /**
     * Creates a Nub from the given properties, and associates it to the current form
     */
    public initNub(props?: Props) {
        const p = props ? props : new Props(this.defaultProperties);
        const p2 = p.clone();
        this.currentNub = Session.getInstance().createSessionNub(p);
        if (this.currentNub instanceof SectionNub) {
            // add new Section as first child, from given nodeType
            p2.setPropValue("calcType", CalculatorType.Section);
            const section = Session.getInstance().createNub(p2);
            this.currentNub.setSection(section as acSection);
        }
    }

    public get currentNub(): Nub {
        return this._currentNub;
    }

    public set currentNub(n: Nub) {
        const nubCalcType = (n.properties as Props).getPropValue("calcType");
        if (this._props["calcType"] !== nubCalcType) {
            throw new Error(
                `Nub ${n.properties["calcType"]} incompatible avec le formulaire ${this._calculatorName} (${this._props["calcType"]})`
            );
        }
        // unsubscribe from old Nub
        if (this._currentNub) {
            this._currentNub.removeObserver(this);
        }
        // replace Nub
        this._currentNub = n;
        // subscribe to new Nub (for result updates)
        this._currentNub.addObserver(this);
    }

    protected deleteNub(sn: Nub) {
        Session.getInstance().deleteNub(sn);
    }

    protected parseOptions(json: {}) {
        const dnt = json["defaultNodeType"];
        this._props["nodeType"] = dnt === undefined ? ComputeNodeType.None : ComputeNodeType[dnt];
        this.helpLink = json["help"];
    }

    protected completeParse(json: {}) {
    }

    public getOption(json: {}, option: string): string {
        if (json["type"] === "options") {
            return json[option];
        }
    }

    public afterParseFieldset(fs: FieldSet) {
    }

    public createFieldset(parent: FormulaireNode, json: {}, data?: {}, nub?: Nub): FieldSet {
        const res: FieldSet = new FieldSet(parent);
        res.setNub(this._currentNub, false);
        this.kids.push(res);
        return res;
    }

    /**
     * @return true si le SessionNub attaché (ou un de ses enfants) a l'uid donné
     * @param uid id à rechercher
     */
    public hasNubId(uid: string): boolean {
        return (this._currentNub && this._currentNub.uid === uid);
    }

    public moveFieldsetUp(fs: FieldSet) {
    }

    public moveFieldsetDown(fs: FieldSet) {
    }

    public removeFieldset(fs: FieldSet) {
    }

    private parse_fieldset(json: {}, nub?: Nub) {
        const fs = this.createFieldset(this, json, undefined, nub);
        fs.parseConfig(json);
        this.afterParseFieldset(fs);
    }

    private parse_template_container(json: {}, templates: any[]) {
        const fsc: FieldsetContainer = new FieldsetContainer(this);
        fsc.parseConfig(json, templates);
        this.kids.push(fsc);
    }

    public parseDependencies(json: {}) {
        // tslint:disable-next-line:forin
        for (const conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            switch (type) {
                // field set
                case "fieldset":
                case "template_container":
                    for (const k of this.kids) {
                        if (k.id === conf["id"]) {
                            k.parseDependencies(conf);
                            break;
                        }
                    }
                    break;

                case "fieldset_template":
                    for (const k of this.kids) {
                        if (k instanceof FieldsetContainer) {
                            k.parseDependencies(conf);
                        }
                    }
                    break;
            }
        }
    }

    /**
     * 1ère passe d'analyse de la configuration
     */
    public preparseConfig(json: {}) {
        this._jsonConfig = json;

        // analyse des options globales
        // il est utile de le faire avant le reste pour les modules de calcul utilisant
        // des sections (id des selects type de section/variable à calculer)

        // tslint:disable-next-line:forin
        for (const conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            if (type === "options") {
                this.parseOptions(conf);
                break;
            }
        }
    }

    /**
     * 2ème passe d'analyse de la configuration
     */
    public parseConfig(json?: {}) {
        if (json !== undefined) {
            this._jsonConfig = json;
        }

        // analyse des éléments du formulaire
        const templates: any[] = [];

        // tslint:disable-next-line:forin
        for (const conf_index in this._jsonConfig) {
            const conf = this._jsonConfig[conf_index];
            const type: string = conf["type"];

            switch (type) {
                case "fieldset":
                    this.parse_fieldset(conf);
                    break;

                case "fieldset_template":
                    templates.push(conf);
                    break;

                // options globales
                case "options":
                    break;

                case "template_container":
                    this.parse_template_container(conf, templates);
                    break;

                default:
                    throw new Error(`type d'objet de module de calcul ${type} non pris en charge`);
            }
        }

        this.completeParse(this._jsonConfig);
        this.parseDependencies(this._jsonConfig);
    }

    public hasParameter(symbol: string): boolean {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter) {
                if (p.symbol === symbol) {
                    return true;
                }
            }
        }
        return false;
    }

    public getParamFromSymbol(symbol: string): NgParameter {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter) {
                if (p.symbol === symbol) {
                    return p;
                }
            }
        }
    }

    public getDisplayedParamFromState(st: ParamRadioConfig): NgParameter {
        for (const p of this.allFormElements) {
            if (p.isDisplayed && p instanceof NgParameter) {
                if (p.radioState === st) {
                    return p;
                }
            }
        }
    }

    public getDisplayedParamListFromState(st: ParamRadioConfig): NgParameter[] {
        const res = [];
        for (const p of this.allFormElements) {
            if (p.isDisplayed && p instanceof NgParameter && p.radioState === st) {
                res.push(p);
            }
        }
        return res;
    }

    public getFieldById(id: string): Field {
        const res = this.getFormulaireNodeById(id);
        if (res instanceof Field) {
            return res;
        }
    }

    public getParameterValue(symbol: string): number {
        for (const fs of this.allFieldsets) {
            const p = fs.getNodeParameter(symbol);
            if (p !== undefined) {
                switch (p.radioState) {
                    case ParamRadioConfig.FIX:
                        return p.getValue();

                    case ParamRadioConfig.VAR:
                    case ParamRadioConfig.CAL:
                        return undefined;
                }
            }
        }

        throw new Error(`Formulaire.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);
    }

    protected notifyReset() {
        this.notifyObservers({
            "action": "resetForm"
        }, this);
    }

    /**
     * Forwards Nub's result updated notification.
     * Used by FormCompute to update results display
     */
    protected notifyNubResultUpdated(sender) {
        this.notifyObservers({
            action: "nubResultUpdated"
        }, sender);
    }

    /**
     * réinitialisation du formulaire suite à un changement d'une valeur, d'une option, ... :
     * effacement des résultats, application des dépendances, ...
     */
    public reset() {
        this.resetResults([]);
        this.applyDependencies();

        // prévenir les composants qu'il faut détecter les changements
        this.notifyReset();
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): any {
        const select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        return select.getValue().value;
    }

    /**
     * retourne le label actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns label courant du select
     */
    public getSelectedLabel(selectFieldId: string): string {
        const select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        return select.getValue().label;
    }

    public applyDependencies() {
        for (const fe of this.topFormElements) {
            fe.applyDependencies();
        }
    }

    public abstract resetResults(visited: string[]);
    public abstract doCompute();
    public abstract get hasResults(): boolean;
    public abstract get results(): CalculatorResults[];

    public updateLocalisation(localisation: StringMap, lang: string) {
        this._specificLocalisation = localisation;
        this._currentLanguage = lang;
        for (const fe of this.topFormElements) {
            fe.updateLocalisation(localisation);
        }
        if (this.hasResults) {
            this.doCompute(); // pour mettre à jour la langue des intitulés de résultats
        }
    }

    public get helpLink(): string {
        return this._helpLink;
    }

    public set helpLink(h: string) {
        this._helpLink = h;
    }

    public isDisplayed(id: string) {
        return (<FormulaireElement>this.getFormulaireNodeById(id)).isDisplayed;
    }

    /**
     * gestion d'un clic sur les radios
     */
    public onRadioClick(info: any) {
        // if mode changed, reset form results
        if (info.oldValueMode !== info.param.valueMode) {
            this.reset();
        }
    }

    /**
     * itère sur tous les FieldSet
     */
    public get allFieldsets(): IterableIterator<FieldSet> {
        return new DeepFieldsetIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement
     */
    public get allFormElements(): IterableIterator<FormulaireElement> {
        return new DeepFormulaireElementIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement de 1er niveau
    */
    private get topFormElements(): IterableIterator<FormulaireElement> {
        return new TopFormulaireElementIterator(this);
    }

    /**
     * MAJ des liens entre paramètres lors de la désérialisation
     */

    private getNthFieldset(n: number): FieldSet {
        let i = 0;
        for (const e of this.topFormElements) {
            if (e instanceof FieldSet) {
                if (i === n) {
                    return e;
                }
                i++;
            }
        }

        return undefined;
    }

    //  interface Observer

    public update(sender: any, data: any) {
        if (sender instanceof Nub) {
            switch (data.action) {
                case "resultUpdated":
                    // forward Nub results update notification to FormCompute objects
                    this.notifyNubResultUpdated(sender);
                    break;
            }
        }
    }
}
