import { PlottableData } from "./plottable-data";
import { PabResults } from "./pab-results";
import { GraphType } from "./graph-type";
import { ServiceFactory } from "../services/service-factory";

export class PlottablePabResults implements PlottableData {

    public graphType: GraphType = GraphType.Scatter;
    public chartX: string;
    public chartY: string;

    protected pabResults: PabResults;

    public constructor(pabResults?: PabResults) {
        if (pabResults) {
            this.setPabResults(pabResults);
        }
        // axes par défaut
        this.chartX = this.chartX || "CLOISON";
        this.chartY = this.chartY || "Z";
    }

    /** reaffect pabResults, for ex. when objet was contructed with empty pabResults */
    public setPabResults(pabResults: PabResults) {
        this.pabResults = pabResults;
    }

    /**
     * Returns the label to display, for an element of getAvailableChartAxis()
     * @param symbol parameter / result symbol (ex: "Q")
     */
    public getChartAxisLabel(symbol: string) {
        return this.pabResults.headers[this.pabResults.columns.indexOf(symbol)];
    }

    /**
     * Returns a list of plottable parameters / result elements, that can be defined
     * as X or Y chart axis
     */
    public getAvailableChartAxis(): string[] {
        return this.pabResults.columns;
    }

    /**
     * Returns the series of values for the required symbol
     * @param symbol parameter / result symbol (ex: "Q")
     */
    public getValuesSeries(symbol: string): any[] {
        const data: string[] = [];
        const pr = this.pabResults;
        const nDigits = ServiceFactory.instance.applicationSetupService.displayDigits;
        const l = this.pabResults.cloisonsResults.length;

        if (this.hasError()) {
            return [];
        }

        switch (symbol) {
            case "CLOISON":
                data.push("");
                for (let i = 0; i < l; i++) {
                    data.push("" + (i + 1));
                }
                data.push(ServiceFactory.instance.i18nService.localizeText("INFO_EXTRARES_LIB_AVAL"));
                break;

            case "DH":
                data.push("");
                for (let i = 0; i < l - 1; i++) {
                    data.push((pr.cloisonsResults[i].vCalc - pr.cloisonsResults[i + 1].vCalc).toFixed(nDigits));
                }
                data.push((pr.cloisonsResults[l - 1].vCalc - pr.cloisonAvalResults.vCalc).toFixed(nDigits));
                data.push((pr.cloisonAvalResults.vCalc - pr.Z2).toFixed(nDigits));
                break;

            case "ZRAM":
                data.push("");
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].getExtraResult("ZRAM");
                    data.push((er !== undefined) ? er.toFixed(nDigits) : "");
                }
                const zrAval = pr.cloisonAvalResults.getExtraResult("ZRAM");
                data.push((zrAval !== undefined) ? zrAval.toFixed(nDigits) : "");
                break;

            case "Q":
                data.push("");
                for (let i = 0; i < l; i++) {
                    data.push(pr.cloisonsQ[i].toFixed(nDigits));
                }
                data.push(pr.cloisonAvalQ.toFixed(nDigits));
                break;

            case "Z":
                for (let i = 0; i < l - 1; i++) {
                    data.push(pr.cloisonsResults[i].vCalc.toFixed(nDigits));
                }
                data.push(pr.cloisonAvalResults.vCalc.toFixed(nDigits));
                data.push(pr.Z2.toFixed(nDigits));
                break;

            case "PV":
                data.push("");
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].getExtraResult("PV");
                    data.push((er !== undefined) ? er.toFixed(nDigits) : "");
                }
                data.push("");
                break;

            case "YMOY":
                data.push("");
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].getExtraResult("YMOY");
                    data.push((er !== undefined) ? er.toFixed(nDigits) : "");
                }
                data.push("");
                break;

            case "ZRB":
                data.push("");
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].getExtraResult("ZRB");
                    data.push((er !== undefined) ? er.toFixed(nDigits) : "");
                }
                data.push("");
                break;

            case "QA":
                data.push("");
                for (let i = 0; i < l; i++) {
                    data.push(pr.bassinsQA[i].toFixed(nDigits));
                }
                data.push("");
                break;
        }

        return data;
    }

    protected hasError(): boolean {
        return this.pabResults.hasError();
    }
}
